//
//  OpenCVBlobTrackPlugIn.m
//  OpenCVBlobTrack
//
//  Created by Jeroen Bouma on 10/21/11.
//  Copyright 2011 iPhonk. All rights reserved.
//
//  Thanks go to mirek for his fine example on how to convert a QCInputImage to IplImage:
//  https://github.com/mirek/quartzcomposer-opencv
//

/* It's highly recommended to use CGL macros instead of changing the current context for plug-ins that perform OpenGL rendering */
#import <OpenGL/CGLMacro.h>
#include <cv.h>
#include <highgui.h>
#include <cvblob.h>


#import "OpenCVBlobTrackPlugIn.h"

#define	kQCPlugIn_Name				@"OpenCV Blob Track"
#define	kQCPlugIn_Description		@"Analyzes the input image with OpenCV and outputs a struct with xy coordinates of the found blobs. The input image can automagically be downsized by 'Size divider' for a performance boost."

using namespace cvb;

@implementation OpenCVBlobTrackPlugIn

@dynamic inputImage, outputBlobCoordinates, inputDivideSizeBy, inputMovementThreshold, inputBlobSizeThreshold;

/*
Here you need to declare the input / output properties as dynamic as Quartz Composer will handle their implementation
@dynamic inputFoo, outputBar;
*/

+ (NSDictionary *)attributes
{
	/*
	Return a dictionary of attributes describing the plug-in (QCPlugInAttributeNameKey, QCPlugInAttributeDescriptionKey...).
	*/
	
	return [NSDictionary dictionaryWithObjectsAndKeys:kQCPlugIn_Name, QCPlugInAttributeNameKey, kQCPlugIn_Description, QCPlugInAttributeDescriptionKey, nil];
}

+ (NSDictionary *)attributesForPropertyPortWithKey:(NSString *)key
{
	/*
	Specify the optional attributes for property based ports (QCPortAttributeNameKey, QCPortAttributeDefaultValueKey...).
	*/
	if([key isEqualToString:@"inputImage"])
        return [NSDictionary dictionaryWithObjectsAndKeys:@"Blob image", QCPortAttributeNameKey, nil];
    if([key isEqualToString:@"inputDivideSizeBy"])
        return [NSDictionary dictionaryWithObjectsAndKeys:
                @"Size divider", QCPortAttributeNameKey, 
                [NSNumber numberWithInt:1], QCPortAttributeDefaultValueKey, 
                [NSNumber numberWithInt:1], QCPortAttributeMinimumValueKey, 
                [NSNumber numberWithInt:50], QCPortAttributeMaximumValueKey, 
                nil];
    if([key isEqualToString:@"inputMovementThreshold"])
        return [NSDictionary dictionaryWithObjectsAndKeys:
                @"Movement threshold", QCPortAttributeNameKey, 
                [NSNumber numberWithInt:2], QCPortAttributeDefaultValueKey, 
                [NSNumber numberWithInt:0], QCPortAttributeMinimumValueKey, 
                [NSNumber numberWithInt:50], QCPortAttributeMaximumValueKey, 
                nil];
    if([key isEqualToString:@"inputBlobSizeThreshold"])
        return [NSDictionary dictionaryWithObjectsAndKeys:
                @"Blob size threshold", QCPortAttributeNameKey, 
                [NSNumber numberWithInt:100], QCPortAttributeDefaultValueKey, 
                [NSNumber numberWithInt:0], QCPortAttributeMinimumValueKey, 
                [NSNumber numberWithInt:500], QCPortAttributeMaximumValueKey,
                nil];
    
    if([key isEqualToString:@"outputBlobCoordinates"])
        return [NSDictionary dictionaryWithObjectsAndKeys:@"Blob coordinates", QCPortAttributeNameKey, nil];
    
	return nil;
}

+ (QCPlugInExecutionMode)executionMode
{
	/*
	Return the execution mode of the plug-in: kQCPlugInExecutionModeProvider, kQCPlugInExecutionModeProcessor, or kQCPlugInExecutionModeConsumer.
	*/
	return kQCPlugInExecutionModeProcessor;
}

+ (QCPlugInTimeMode)timeMode
{
	/*
	Return the time dependency mode of the plug-in: kQCPlugInTimeModeNone, kQCPlugInTimeModeIdle or kQCPlugInTimeModeTimeBase.
	*/
	
	return kQCPlugInTimeModeNone;
}

- (id)init
{
	self = [super init];
	if (self) {
		/*
		Allocate any permanent resource required by the plug-in.
		*/
        
        previousBlobs = [[NSMutableDictionary alloc] initWithCapacity:4];
	}
	
	return self;
}

- (void)finalize
{
	/*
	Release any non garbage collected resources created in -init.
	*/
	
	[super finalize];
}

- (void)dealloc
{
	/*
	Release any resources created in -init.
	*/
	
    [previousBlobs release];
    
	[super dealloc];
}


@end

@implementation OpenCVBlobTrackPlugIn (Execution)

- (BOOL)startExecution:(id <QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when rendering of the composition starts: perform any required setup for the plug-in.
	Return NO in case of fatal failure (this will prevent rendering of the composition to start).
	*/
	
	return YES;
}

- (void)enableExecution:(id <QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when the plug-in instance starts being used by Quartz Composer.
	*/
    
}

// Quartz Composer supports:
// * ARGB8 (8-bit alpha, red, green, blue)
// * BGRA8 (8-bit blue, green, red, and alpha)
// * RGBAf (floating-point, red, green, blue, alpha)
// * I8    (8-bit intensity), and
// * If    (floating-point intensity).
+ (IplImage *) createIplImageHeaderWithInputImageSource: (id<QCPlugInInputImageSource>) image {
    IplImage *result = NULL;
    if (image != nil) {
        CvSize size = cvSize([image imageBounds].size.width, [image imageBounds].size.height);
        const char *colorModel;
        const char *channelSeq;
        int depth;
        int channels;
        if ([image bufferPixelFormat] == QCPlugInPixelFormatARGB8) {
            depth = IPL_DEPTH_8U;
            channels = 4;
            colorModel = (char *)"RGBA";
            channelSeq = (char *)"ARGB";
        } else if ([image bufferPixelFormat] == QCPlugInPixelFormatBGRA8) {
            depth = IPL_DEPTH_8U;
            channels = 4;
            colorModel = (char *)"RGBA";
            channelSeq = (char *)"BGRA";
        } else if ([image bufferPixelFormat] == QCPlugInPixelFormatRGBAf) {
            depth = IPL_DEPTH_32F;
            channels = 4;
            colorModel = (char *)"RGBA";
            channelSeq = (char *)"RGBA";
        } else if ([image bufferPixelFormat] == QCPlugInPixelFormatI8) {
            depth = IPL_DEPTH_8U;
            channels = 1;
            colorModel = (char *)"GREY";
            channelSeq = (char *)"GREY";
        } else if ([image bufferPixelFormat] == QCPlugInPixelFormatIf) {
            depth = IPL_DEPTH_32F;
            channels = 1;
            colorModel = (char *)"GREY";
            channelSeq = (char *)"GREY";
        } else {
            NSLog(@"* Unsupported image type %@", [image bufferPixelFormat]);
            assert(false);
        }
        result = cvCreateImageHeader(size, depth, channels);
        result->imageData = (char *)[image bufferBaseAddress];
        strcpy(result->colorModel, colorModel);
        strcpy(result->channelSeq, channelSeq);
    }
    return result;
}

- (BOOL)execute:(id <QCPlugInContext>)context atTime:(NSTimeInterval)time withArguments:(NSDictionary *)arguments
{
	/*
	Called by Quartz Composer whenever the plug-in instance needs to execute.
	Only read from the plug-in inputs and produce a result (by writing to the plug-in outputs or rendering to the destination OpenGL context) within that method and nowhere else.
	Return NO in case of failure during the execution (this will prevent rendering of the current frame to complete).
	
	The OpenGL context for rendering can be accessed and defined for CGL macros using:
	CGLContextObj cgl_ctx = [context CGLContextObj];
	*/
    
    id<QCPlugInInputImageSource> inputImage = self.inputImage;
    
    // no image? abort
    if(inputImage == nil)
    {
        self.outputBlobCoordinates = nil;
        [context logMessage:@"no input image for blob detection"];
        return YES;
    }

    BOOL result = NO;
    NSString *pixelFormat;
    CGColorSpaceRef colorSpace = [inputImage imageColorSpace];
    if (CGColorSpaceGetModel(colorSpace) == kCGColorSpaceModelMonochrome) {
        pixelFormat = QCPlugInPixelFormatI8;
    } else if (CGColorSpaceGetModel(colorSpace) == kCGColorSpaceModelRGB) {
#if __BIG_ENDIAN__
        pixelFormat = QCPlugInPixelFormatARGB8;
#else
        pixelFormat = QCPlugInPixelFormatBGRA8;
#endif
    } else {
        [context logMessage: @"Unable to get proper pixel format to lock buffer representation"];
        return NO;
    }
    
    
    if (![inputImage lockBufferRepresentationWithPixelFormat: pixelFormat
                                                   colorSpace: [inputImage imageColorSpace]
                                                    forBounds: [inputImage imageBounds]]) {
        [context logMessage:
            @"Can't lock buffer representation with pixel format %@, color space %@ and image bounds %@",
            [inputImage bufferPixelFormat],
            [inputImage imageColorSpace],
            [inputImage imageBounds]];
        result = NO;
    } 
    else 
    {
        IplImage *sourceImg = [OpenCVBlobTrackPlugIn createIplImageHeaderWithInputImageSource:inputImage];
        
        // resize image, divide size by inputDivideSizeBy
        float divideSizeBy = (float) self.inputDivideSizeBy;
        if(divideSizeBy < 1 || divideSizeBy > 100)
            divideSizeBy = 1;
        
        IplImage* resizedImg;
        BOOL imageNeededResize = NO;
        if(divideSizeBy == 1)
            resizedImg = sourceImg;
        else
        {
            resizedImg = cvCreateImage(cvSize(sourceImg->width/divideSizeBy, sourceImg->height/divideSizeBy), sourceImg->depth, sourceImg->nChannels);
            cvResize(sourceImg, resizedImg);
            imageNeededResize = YES;
        }
        
        IplImage* blackWhiteImg = nil;
        BOOL imageNeededBlackWhitening = NO;
        
        // process image to binary format (if required/possible)
        if(resizedImg->nChannels == 4)
        {
            imageNeededBlackWhitening = YES;
            
            blackWhiteImg = cvCreateImage(cvGetSize(resizedImg), IPL_DEPTH_8U, 1);
            cvCvtColor(resizedImg, blackWhiteImg, CV_RGBA2GRAY);
            cvThreshold(blackWhiteImg, blackWhiteImg, 150, 255, CV_THRESH_BINARY);
        }
        else if(resizedImg->nChannels == 3)
        {
            imageNeededBlackWhitening = YES;
            
            blackWhiteImg = cvCreateImage(cvGetSize(resizedImg), IPL_DEPTH_8U, 1);
            cvCvtColor(resizedImg, blackWhiteImg, CV_RGB2GRAY);
            cvThreshold(blackWhiteImg, blackWhiteImg, 150, 255, CV_THRESH_BINARY);
        }
        else if(resizedImg->nChannels == 1)
        {
            blackWhiteImg = resizedImg;
        }
        
        // could a grayscale image be generated?
        if(blackWhiteImg)
        {
            // find blobs
            IplImage* labelImg = cvCreateImage(cvGetSize(blackWhiteImg), IPL_DEPTH_LABEL, 1);
            CvBlobs blobs;
            unsigned int result = cvLabel(blackWhiteImg, labelImg, blobs);
            
            // make output struct out of blob data
            NSMutableDictionary* output = [[NSMutableDictionary alloc] initWithCapacity:4];
            double movementThreshold = self.inputMovementThreshold;
            double blobSizeThreshold = self.inputBlobSizeThreshold;
            
            for (CvBlobs::const_iterator it=blobs.begin(); it!=blobs.end(); ++it)
            {
                //printf("Blob x: %lf, y:%lf \n", it->second->centroid.x, it->second->centroid.y);
                
                NSString* labelString = [NSString stringWithFormat:@"%d", it->second->label];
                BlobData* blobData = [previousBlobs objectForKey:labelString];
                
                double currentBlobX = it->second->centroid.x * divideSizeBy;
                double currentBlobY = it->second->centroid.y * divideSizeBy;
                double currentBlobArea = it->second->area * divideSizeBy;
                
                // check for minimum size, ignore blob if too small
                if(currentBlobArea > blobSizeThreshold)
                {
                    
                    // no previous blob for this label
                    if(!blobData)
                    {
                        // create data object and store it
                        blobData = [[BlobData alloc] initWithX:it->second->centroid.x y:it->second->centroid.y];
                        [previousBlobs setObject:blobData forKey:labelString];
                        [blobData release];
                    }
                    else
                    {
                        // blob label found in previous blobs, calculate difference between current and previous
                        double deltaX = fabs(currentBlobX - blobData.x);
                        double deltaY = fabs(currentBlobY - blobData.y);
                        
                        // difference larger then threshold?
                        if(deltaX > movementThreshold
                           || deltaY > movementThreshold)
                        {
                            // blob moved significantly, use new coords
                            blobData.x = currentBlobX;
                            blobData.y = currentBlobY;
                        }
                        else
                        {
                            // blob did not move enough, use old coords
                        }
                    }
                    
                    NSDictionary* blobCoords = [NSDictionary dictionaryWithObjectsAndKeys:
                                                [NSNumber numberWithDouble:blobData.x], @"x", 
                                                [NSNumber numberWithDouble:blobData.y], @"y", 
                                                [NSNumber numberWithDouble:currentBlobArea], @"size", 
                                                nil];
                    [output setObject:blobCoords forKey:labelString];
                }
            }
            
            // free label image
            cvReleaseImage(&labelImg);
            cvReleaseBlobs(blobs);
            
            self.outputBlobCoordinates = output;
            [output release];
            
        }
        else
        {
            [context logMessage:@"can't process input image: not supported. channels: %d, sequence: %s", resizedImg->nChannels, resizedImg->channelSeq];
        }
        
        // release opencv image header/data
        cvReleaseImageHeader(&sourceImg);
        
        if(imageNeededResize)
            cvReleaseImage(&resizedImg);
        if(imageNeededBlackWhitening)
            cvReleaseImage(&blackWhiteImg);
        
        [self.inputImage unlockBufferRepresentation];
    }
    
    return YES;
}

- (void)disableExecution:(id <QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when the plug-in instance stops being used by Quartz Composer.
	*/
}

- (void)stopExecution:(id <QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when rendering of the composition stops: perform any required cleanup for the plug-in.
	*/
}

@end


////////////////////////

@implementation BlobData
@synthesize x, y;

- (id) initWithX:(double)xx y:(double)yy
{
    if((self = [super init]))
    {
        x = xx;
        y = yy;
    }
    
    return self;
}

@end

